#$Id: Bootpd.pm,v 1.4 2000/04/03 06:47:34 ht8 Exp $

=head1 NAME

Bootpd - partial implementation of BOOTP Daemon

=head1 SYNOPSIS

  use Bootpd;

  non-blocking:
  my $b = new Bootpd (bootptab     => "/etc/bootptab", 
                      non_blocking => 1,
                      verbose      => 1);
  my $status = $b->recv;
  select (...); #aditional code and parameters required.
  $b->reply;

  blocking:
  my $b = new Bootpd (bootptab     => "/etc/bootptab");

  my $status = $b->recv; #execution blocks here. Continues after a request
                         #was received



=head1 DESCRIPTION

This Module implements parts of RFC 951. As of now, only 
the subset required to boot a DEC LPS17 printer is implemented.

It should not be used as a general bootbd, as there were 
no precautions taken to make this a re-entrant implementation, 
e.g. a second boot request will overwrite the first one.

Any improvmenet of this module is highly appreciated...


=head1 CONFIG FILE

The format is as the one of bootptab in Red Hat Linux. 
With one exception: On Microsoft platforms, the delemiter is ";"
instead of ":". Otherwise the colon after the drive letter is 
mistaken as a delemiter.

e.g. Microsoft:
  lps17; ht=ethernet; ha=08002b955ef9; ip=160.59.218.20;\
  hd=/tftpboot; bf=lps17.sys; sm=255.255.255.0;  gw=160.59.218.20;

e.g. other platforms
  lps17: ht=ethernet: ha=08002b955ef9: ip=160.59.218.20:\
  hd=/tftpboot: bf=lps17.sys: sm=255.255.255.0:  gw=160.59.218.20:

            bf   Bootfile
            gw   Gateway address list
            ha   Host hardware address
            hd   Bootfile home directory
            ht   Host hardware type (see Assigned Numbers RFC)
            ip   Host IP address
            sm   Host subnet mask


The following parameters are allowed in Red Hat Linux, but not supported 
by this re-implementation:

            bs   Bootfile size in 512-octet blocks
            cs   Cookie server address list
            df   Merit dump file
            dn   Domain name
            ds   Domain name server address list
            ef   Extension file
            hn   Send client'' s hostname to client
            im   Impress server address list
            lg   Log server address list
            lp   LPR server address list
            ns   IEN-116 name server address list
            nt   NTP (time) Server (RFC 1129)
            ra   Reply address override
            rl   Resource location protocol server address list
            rp   Root path to mount as root
            sa   TFTP server address client should use
            sw   Swap server address
            tc   Table continuation (points to similar "template" host entry)
            td   TFTP  root  directory  used  by  "secure" TFTP servers
            to   Time offset in seconds from UTC
            ts   Time server address list
            vm   Vendor magic cookie selector
            yd   YP (NIS) domain name
            ys   YP (NIS) server address


=cut


package Bootpd;
use strict;
use File::Basename;
use 5.004;
use Getopt::Std;
use Carp qw(carp);
use Sys::Hostname;
use IO::Socket;

my $redir= "2>&1";
if ($^O eq "MSWin32") {
  ## Win9x systems do not support redirection, so we need to take care of
  ## this. Check if redirection works as expected, if not set $redir to ""
  unlink "&1" if -f "&1";
  open CMD, "dir 2>&1 |" or die "dir failed: $!";
  my @resp = <CMD>;
  close CMD or die "dir failed: $!";
  if (-f "&1") {
    $redir = "";
    unlink "&1";
  }
}

## Format required to read/write a BOOTP request/response 
##
## pack is using the following syntax:
## C: An unsigned char value.
## L: An unsigned long value,  _exactly_ 32 bits
## S: An unsigned short value, _exactly_ 16 bits
## a: string with arbitrary binary data, will be null padded.
## 
my $format = "CCCCLSSLC4C4C4a16Z64Z128a64";

=head1 METHODS

=head2 new (bootptab=>config_file [,verbose=>0|1|2] [, non_blocking=>0 | 1])


=cut

sub new {
  my $class             = shift;
  my %pars              = @_;
  my $self              = {};
  my ($key, $val);
  my $errors            = 0;
  local $_;

  $self->{dbg}          = 0;
  $self->{non_blocking} = 0;
  while (($key, $val) = each %pars) {
    if    ($key =~ /^bootptab$/i) {
      $self->{bootptab} = $val;
    }
    elsif ($key =~ /^non_blocking$/i) {
      $self->{non_blocking} = $val;
    }
    elsif ($key =~ /^verbose$/i) {
      $self->{dbg} = $val;
    }
    else {
      carp "Parameter \"$key\"=>\"$val\" is not supported";
      $errors += 1;
    }
  }
  return undef if $errors;

  unless ($self->{sock_bootps} = IO::Socket::INET->new
          (LocalPort=> 'bootps(67)',
           Proto    => 'udp')) {
    carp "Failed to establish bootps socket: $!";
    return undef;
  }


  if ($self->{non_blocking}){
    use Fcntl;

    my $flags = 0;
    fcntl($self->{sock_bootps}, F_GETFL, $flags)
      or die "Couldn't get flags for HANDLE : $!\n";
    $flags |= O_NONBLOCK;
    fcntl($self->{sock_bootps}, F_SETFL, $flags)
      or carp "Couldn't put bootps handle into non-blocking mode: $!\n";
  }

  bless $self, $class;
}


=head2 recv 

If in normal mode, wait for a boot request. The reply will be sent 
as well. This is done by calling the reply method. The status of the 
reply method is returned.

If on non-blocking mode, return to caller immediately.

=cut 

sub recv {
  my $self = shift;

  $self->{sock_bootps}->recv($self->{request}{data}, 1024);

  ##
  ## only reply if in blocking mode. If in non-blocking mode, there is no 
  ## response yet.
  ##
  unless ($self->{non_blocking}) {return ($self->reply)};
}


=head2 reply

Use this method only if in non-blocking operation. If in blocking mode,
this "reply" is called from within recv.

Return values:

=over 5

=item undef 

if anything goes wrong.

=item "no entry" 

if the clients hw address could not be found in the config file. 

=item "reply sent" 

if a bootp response was sent to the requesting client.

=back

=cut

sub reply {
  my $self = shift;

  $self->unpack_req;

  return "no entry" if ! $self->get_entry;
  unless ($self->pack_reply) {
    carp "packing reply failed\n";
    return undef;
  }

  unless ($self->bootp_reply) {
    carp "could not send reply to $self->{request}{hwa_arp}\n";
    return undef;
  }

  return "reply sent";
#  if ($self->get_entry) {
#    if ($self->pack_reply) {
#      $self->bootp_reply;
#    }
#  }
}


sub bootp_reply {
  my $self = shift;

  unless ($self->{sock_bootpc} = IO::Socket::INET->new
          (PeerPort => 'bootpc(68)',
           Proto    => 'udp',
           PeerAddr => $self->{client}{ip})) {
    carp "Failed to establish bootpc socket: $!";
    return undef;
  }


  ## issue arp -s command. This defines an entry so the kernel is able 
  ## to send the udp package to the boot client.

  my @cmds = ();
  $cmds[0] = "arp -s $self->{client}{ip} $self->{request}{hwa_arp}";
  $cmds[1] = "arp -d $self->{client}{ip}";

  foreach my $cmd (@cmds) {
    unless (open CMD, "$cmd $redir |") {
      carp "spawning of command $cmd failed: $!";
      return undef;
    }
    my $resp = <CMD>;
    unless (close CMD) {
      carp "command $cmd failed: $resp";
      return undef;
    }
    &log("command \"$cmd\" successfully completed") if $self->{dbg};
    if ($cmd =~ /^arp -s/) {
      &log("Sending bootp reply") if $self->{dbg};
      unless ($self->{sock_bootpc}->send($self->{reply}{data})) {
        carp "send failed\n";
        return undef;
      }
    }
  }
  &log("Bootp response successfully sent to $self->{request}{hwa_arp}") 
    if $self->{dbg};
  
  close $self->{sock_bootpc};
}


## Read config file, search for given hw address. Return parameters for given
## hw address if found, undef otherwise.
sub get_entry {
  my $self = shift;
  my $status = undef;

  foreach my $k (qw (bf hd gw ht ip sm ha)) {
    $self->{client}{$k} = undef;
  }

  unless (exists $self->{bootptab}) {
    carp "path to config file was not given\n";
    return undef;
  }
  unless (open BT, $self->{bootptab}) {
    carp "failed to open $self->{bootptab}: $!\n";
    return undef;
  }

  ## Loop to strip of comments, concat lines ending with \
  my $line;
  my $del = ($^O eq "MSWin32") ? ";" : ":";
 read_cfg:
  while (<BT>) {
    s/\#.*$//; #weed out comments
    s/\s+/ /g; #replace white space with single blank
    s/^ $//;   #if there is only one blank, make line empty
    next read_cfg if /^$/;
    $line = (/^[a-z0-9-_]+$del\s+(.*)/i) ? $1 : $line . $_;
    if ($line =~ /(.*)\\\s*$/) {
      $line = $1;
      next read_cfg;
    }

    print "Working on \"$line\"\n" if $self->{dbg} > 1;
    if ($line =~ /ha=$self->{request}{hwa_str}$del/i) {
      &log("Found ha=$self->{request}{hwa_str} in $self->{bootptab}");
      $status = 1;
    element:
      foreach my $elem (split (/\s*$del/, $line)) {
        my ($k, $v) = ($elem =~ /\s*(..)=(.*)/);
        next element if $elem =~ /^\s+$/;
#        print "elem=\"$elem\"\n";
        if (defined $k and $k ne "" and exists $self->{client}{$k}) {
          $self->{client}{$k} = $v;
        }
        else {
          carp "Parameter \"$k\" is not supported";
        }
      }
      last read_cfg;
    }
  }
  close BT;

  if ($self->{dbg}) {
    if ($status) {
      &log("IP parameters for requesting client");
      &log("Ethernet Address: $self->{request}{hwa_arp}");
      &log("IP Address      : $self->{client}{ip}");
      &log("Subnet Mask     : $self->{client}{sm}");
      &log("default Gateway : $self->{client}{gw}");
      &log("boot file       : $self->{client}{hd}/$self->{client}{bf}");
      &log(sprintf "completed at    : %s", scalar localtime);
    }
    else {
      &log(sprintf "There was no match for \"%s\" in file \"%s\"", 
      $self->{request}{hwa_str}, $self->{bootptab});
    }
  }

  return $status;
}

sub unpack_req {
  my $self = shift;
  my @uu; #to store unused values

  $self->{request}{op} = unpack "C", $self->{request}{data};
  unless ($self->{request}{op} == 1) {
    carp "this is not a request: op=$self->{request}{op}\n";
    return undef;
  }

#  printf "The received request has a overall length of %D\n", 
#  length $self->{request}{data};

  ($self->{request}{op},                     # C
   $self->{request}{htype},                  # C
   $self->{request}{hlen},                   # C
   $uu[0], #$self->{request}{hops},          # C
   $self->{request}{xid},                    # L
   $self->{request}{secs},                   # S
   $uu[1],                                   # S
   $self->{request}{ciaddr},                 # L
   $uu[2], $uu[3], $uu[4], $uu[5], #$self->{request}{yiaddr}, # C4
   $uu[6], $uu[7], $uu[8], $uu[9], #$self->{request}{siaddr}, # C4
   $uu[10], $uu[11], $uu[12], $uu[13], #$self->{request}{giaddr}, # C4
   $self->{request}{chaddr},                 # a16
   $uu[14], #$self->{request}{sname},
   $uu[15], #$self->{request}{file},
   $self->{request}{vend} )= #
     unpack $format, $self->{request}{data};

  unless ($self->{request}{htype} == 1) {
    carp "HW address type ($self->{request}{htype}) not supported";
    carp "always use ethernet (1)";
    return undef;
  }

  unless ($self->{request}{hlen} == 6) {
    carp "length must be 6 bytes";
    return undef;
  }

  $self->{request}{hwa_str} = uc sprintf "%02x%02x%02x%02x%02x%02x", 
  unpack ("C6", $self->{request}{chaddr});

  ## Prepare arp -s command (to reach bootp client 

  ## Windoze is using the form 08-00-2b..., unix 08:00:2b...
  my $d = ($^O =~ /win/i) ? "-" : ":"; 
  $self->{request}{hwa_arp} = uc sprintf "%02x$d%02x$d%02x$d%02x$d%02x$d%02x", 
  unpack ("C6", $self->{request}{chaddr});

  $self->{request}{magic} = substr $self->{request}{vend},0,4;

  &log("received request from $self->{request}{hwa_str}") if $self->{dbg};
}



sub pack_reply {
  my $self = shift;

  $self->{reply}{data} = undef;

  unless ($self->{client}{ht} eq "ethernet") {
    carp "hardware address type \"$self->{client}{ht}\" is not supported";
    return undef;
  }

  my $vend = pack "a4", $self->{request}{magic};

  ## I don't know if the order is important. But I try to be tidy.
  foreach my $option (qw (sm gw)) {
    if (exists $self->{client}{$option} and defined $self->{client}{$option}) {
      if    ($option eq "sm") {
        my @sm   = split (/\./, $self->{client}{sm});
        $vend .= pack "C6", 1,4, @sm;
      }
      elsif ($option eq "gw") {
        my @gw = split (/\./, $self->{client}{gw});
        $vend .= pack "C6", 3,4, @gw;
      }
      else {
        carp "unsupported option \"$option\"";
        return undef;
      }
    }
    else {
      carp "option \"$option\" not used";
    }
  }
  $vend .= pack "C1", 0xff;



  $self->{reply}{data} = pack $format,
  2,         #C    BOOTREPLY
  1,         #C    HW address type. 1 = ethernet
  6,         #C    HW address length
  0,         #C    hops. Used with boot gateways 
             #L    transaction id, as given by client
  $self->{request}{xid},
  0,         #S    secs since client's first request. Ignored by daemon
  0,         #S    unused
  0,         #L    client IP address, as given by client. reset in reponse
             #C4   client's IP address, as assigned by bootpd
  split (/\./,$self->{client}{ip}),

             #C4   server IP address
  unpack ("C4", gethostbyname hostname),

  0,0,0,0,   #C4   gateway IP address (not supported)

             #a16  Client's hw address as given by client 
  substr ($self->{request}{data},28,6),
             
  hostname,  #Z64  optional server name, zero terminated
             #Z128 boot file name. fully qualified directory-pathname
  $self->{client}{hd} . "/" . $self->{client}{bf},

  $vend;     #a64 optional vendor specific data

}


sub log {
  my ($txt) = @_;
  my $date  = localtime;
  ($date)   = ($date =~ /^... (...\s+\d{1,2} ..:..:..)/);
  printf "$date: %s\n", $txt;
}


=head1 AUTHOR

Beat.Hess@cyberdude.com

=cut





1;
