#!/usr/bin/perl -w

#$Id: lpadm.pl,v 1.5 2000/04/04 18:15:59 ht8 Exp $

=head1 NAME

lpadm - downloads files to DECs LPS17 PrintServer

=head1 SYNOPSIS 

use lpadm -h for a quick overview
use lpadm -H for the gory details

=head1 DESCRIPTION

As there is no supported way to boot a DEC Printserver 17 (aka lps17)
from a Linux or NT4 system, I faced a challange.

=head1 LICENSE

This file may be distributed under the same license as perl.

=head1 BUGS

Cannot put sockets into non-blocking mode on MSWin32 systems. 
Therefore there is no way to detect if the printer was powered down. 

OSes supporting fcntl are able to run non-blocking and will therefore 
be able to detect this condition.

=head1 AUTHOR

Beat.Hess@cyberdude.com

=cut


use strict;
use 5.005;
use Getopt::Std;
use File::Basename;
use lib dirname($0);
use IO::Socket;
use IO::Select;
use Bootpd;
use Tftpd;
use Pod::Text;
use Fcntl;
use POSIX qw(:errno_h setsid);

my %opts;
my @fh     = (); #Contains all filehandles for files to be downloaded
my @fn     = (); #Contains their names
$SIG{TERM} = \&exit_handler;
$SIG{QUIT} = \&exit_handler;
$SIG{INT}  = \&exit_handler;
$SIG{PIPE} = 'IGNORE';
getopts ('Hhvd:a:bc:trl:', \%opts) or &usage;

&podthis if exists $opts{H}; #Generate nice on-line documentation
&usage if !exists $opts{d} or exists $opts{h} or $#ARGV != 0;
if (exists $opts{a}) {
  open  ACC, ">>$opts{a}" or die "Open Accounting $opts{a}: $!";
  close ACC;
}

my $distro   = $opts{d};
my $detach   = exists $opts{r};
my $bootpd   = ! exists $opts{b};
my $debug    = (exists $opts{v}) ? 1 : 0;
   $debug    = (exists $opts{t}) ? 2 : $debug;
my $timeout  = 10;
my $nonblock = 1;
my $startpg  = exists $opts{s};
my $lp       = $ARGV[0];
my %loadf    = ();
my ($Handle, $Length, $Type, $File, $Log_txt, $Seq, $Port) = (undef) x 7;
my $polled   = 0;
my $s        = new IO::Select->new();
$|           = 1;

if ($^O eq "MSWin32") {
  &log("WARNING - cannot detect power down of lps17");
  $nonblock = 0;
}

if ($detach) {
  my $pid = fork;
  exit if $pid;
  die "fork failed: $!" unless defined($pid);
  &POSIX::setsid() or die "Can't start a new session: $!";
  chdir '/'            or die "unable to chdir to /: $!";
  umask 0              or die "umask: $!";
  close STDIN          or die "failed to close STDIN: $!";
}

if ($detach or exists $opts{l}) {
  my $outfile = (exists $opts{l}) ? $opts{l} : $0 . ".log";
  open(STDOUT, ">$outfile") or die "Can't redirect stdout to $outfile: $!";
  open(STDERR, ">&STDOUT")  or die "Can't dup stdout";
  select(STDERR); $| = 1;     # make unbuffered
  select(STDOUT); $| = 1;     # make unbuffered
  chmod 0644, $outfile or die "chmod $outfile: $!";
}

my ($bpd, $tfd);
if ($bootpd) {
  &usage if !exists $opts{c};
  &log("Setting up bootpd") if $debug;
  $bpd = new Bootpd (bootptab     => $opts{c}, 
                     non_blocking => $nonblock,
                     verbose      => $debug) or die;

  $tfd = new Tftpd (top_dir       => $distro,
                    non_blocking  => $nonblock,
                    verbose       => $debug) or die;
  $s->add($tfd->{sock_tftp}, $bpd->{sock_bootps}); #not quite OO whant's it..
}
else {
  &log("bootps and tftp turned off upon user request");
}
&mk_printercfg;
&long_to_short(\%loadf);



my $Socket = undef;
my $Event  = "";
my $State  = &dispatch("lp_down", "idle");

##
## Enter infinite main loop
##
my @ready;
my $tftp_inprogress = 0;
while (1) {
  @ready = $s->can_read($timeout); # The parameter is timeout in seconds
  if ($#ready == -1) {
    ## select timed out. This happens in idle or TFTP state respectively
    &get_event("timeout");
    $State = &dispatch ($Event, $State);
    if ($tftp_inprogress) {
      $tftp_inprogress = 0;
      &log("TFTP timed out");
      $tfd->disconnect;
    }
  }
  else {
    foreach my $fh (@ready) {
      if ($fh == $bpd->{sock_bootps}) {
        &log("Received bootp request") if $debug;
        $bpd->recv;
        $bpd->reply;
     }
      elsif (defined $Socket and $fh == $Socket) { # Printer sent some message
        while (&get_event("lps_msg") ne "EAGAIN") {
          $State = &dispatch ($Event, $State);
        }
      }
      elsif ($fh == $tfd->{sock_tftp}) {
        my $tftp_stat = $tfd->recv;
        $tftp_inprogress = (!defined $tftp_stat or 
                            $tftp_stat eq "Xfer complete") ? 0 : 1;
        &log("TFTP xfer successfully completed") if ! $tftp_inprogress;
      }
      else {
        die "unknown file handle";
      }
    }
  }
  if ($#ready >= 0) {
    ## select did not time out. Find out which socket was receiving something
    
  }
}



sub dispatch {
  my ($event, $state) = @_;

  $event    = "erred" if $event eq "";
  my %table = 
  ## state         event       action         new_state
    (idle      => {lp_up   => [\&connect,     "connected"],
                   erred   => [\&nop,         "idle"],
                   lp_down => [\&nop,         "idle"],
                   snd_date=> [\&error,       "idle"],
                   fh_req  => [\&error,       "idle"],
                   data_req=> [\&error,       "idle"],
                   snd_high=> [\&error,       "idle"],
                   fh_close=> [\&error,       "idle"],
                   log_msg => [\&error,       "idle"],
                   acc_msg => [\&error,       "idle"],
#                   timeout => [\&
                   ignore  => [\&error,       "idle"]},
     
     connected => {lp_up   => [\&nop,         "connected"],
                   erred   => [\&disconnect,  "idle"],
                   lp_down => [\&disconnect,  "idle"],
                   snd_date=> [\&send_time,   "connected"],
                   fh_req  => [\&prep_fh,     "connected"],
                   data_req=> [\&send_data,   "connected"],
                   snd_high=> [\&send_cfg,    "connected"],
                   fh_close=> [\&close_fh,    "connected"],
                   log_msg => [\&log_msg,     "connected"],
                   acc_msg => [\&acc_msg,     "connected"],
                   ignore  => [\&nop,         "connected"]}
     );

  ## some sanity checks before we go any further...
  die "inexistant state \"$state\"" if !exists $table{$state};
  die "inexistant event \"$event\"" if !exists $table{$state}{$event};

  my $new_state = $table{$state}{$event}[1];
  die "inexistant new state \"$new_state\"" if !exists $table{$new_state};

  my $action = $table{$state}{$event}[0];
#  &log ("State: $state Event: $event New State: $new_state");
  &$action;
  return $new_state;
}

sub connect {
  &log("Connecting to $lp");
  $Socket = IO::Socket::INET->new(Proto    => "tcp",
                                  PeerAddr => $lp,
                                  PeerPort => "170",
                                  Timeout  => 120)
    or die "cannot connect to port 170 at $lp";

  if ($nonblock) {
    my $flags = 0;
    fcntl($Socket, F_GETFL, $flags)
      or die "Couldn't get flags for HANDLE : $!\n";
    $flags |= O_NONBLOCK;
    fcntl($Socket, F_SETFL, $flags)
      or die "Couldn't put lps17 handle into non-blocking mode: $!\n";
  }
  $s->add($Socket);
  use Sys::Hostname;
  
  my $hostname = hostname();
  my $accountg = exists $opts{a} ? 1 : 0;
  
  my $msg = "PASSWORD=lpsv51\001";
  $msg   .= "HOST=$hostname\001";
  $msg   .= "CFREAD=1\001";
  $msg   .= "FILEIO=1\001";
  $msg   .= "ACCOUNT=$accountg\001";
  $msg   .= "ERRLOG=1\001";
  $msg   .= "PRINTERHOST=$lp\001";
  $msg   .= "CLIENTID=5.1\001";
  $msg   .= "PROTOCOL=V2.0\001";
  $msg   .= "KEEPALIVE=2";#

#  printf $Socket "\00241 1 %d %s", length $msg, $msg or die "write: $!";
    $Socket->send(sprintf ("\00241 1 %d %s", length $msg, $msg)) or 
      die "write: $!";
}

sub poll_lp {
  use Net::Ping;
  my $p  = Net::Ping->new("icmp");
  my $pl = 5;
  my $log= 0;
  my $Event;

  if    ($polled < $pl ) { $log = 1; }
  elsif ($polled == $pl) { &log("I stop logging polling attempts now"); }
  &log("polling printer $lp...") if $log;
  if ($p->ping($lp,1)) {
    &log("$lp alive.") if $log;
    $Event = "lp_up";
  }
  else {
    $Event = "lp_down";
    &log("$lp not reachable. Sleeping $timeout secs.") if $debug and $log; 
  }
  $p->close();
  $polled += 1;
  return $Event;
}

sub send_time {
  my $now    = time;
  my @time_a = localtime ($now);
  my $time_s = localtime ($now);
  my ($month)= ($time_s =~ /^... (...) \d{2} \d{2}:\d{2}:\d{2}/);
  my $ts = sprintf "%02d-%s-%04d %02d:%02d:%02d", $time_a[3], uc $month,
  1900+$time_a[5],$time_a[2],$time_a[1], $time_a[0];
  &log("sending time \"$ts\"");

  printf $Socket "\002101 $Seq %d %s", length $ts, "$ts";
}

sub send_data {
  my ($stat, $data, $header, $info);

  if (eof $fh[$Handle]) {
    $info   = "Download complete";
     &log("$info") if $debug;
    $header = "RETURN=0\001DATA=";
    $data   = " ";
  }
  elsif ($Type eq "r") {
    $info   = "Sending data of type r";
    $data   = readline ($fh[$Handle]); #<CNF>;
    $header = sprintf "RETURN=%d\001DATA=", length $data;
    ## Keep the implementation silly. 
    die "the line exceeds character limit $Length" if length $data > $Length;
  }
  elsif ($Type eq "rb") {
    $info   = "Sending data of type rb";
    $stat   = read $fh[$Handle], $data, $Length;
    die "problem reading file: $!" if !defined $stat;
    die "EOF should never be found here" if $stat == 0;
    $header = "RETURN=$stat\001DATA=";
  }
  else {
    die "Cannot handle files of type \"$Type\"";
  }

  &log("$info") if $debug > 1;
  my $hdr = sprintf "\002101 $Seq %d ", length($header) + length($data);
  print $Socket $hdr . $header . $data or die "write failed: $!";
}


sub send_cfg {
  my %portmap = (1=> 1024, 2 => 1025, 3 => 1026, 0 => 1027);
  &log ("Sending \"$fn[$Handle]\" to port $portmap{$Port % 4}");

  my $Channel = IO::Socket::INET->new(Proto    => "tcp",
                                      PeerAddr => $lp,
                                      PeerPort => $portmap{$Port % 4})
    or die "cannot connect to port $portmap{$Port} at $lp";

  my $buffer;
  while (read $fh[$Handle], $buffer, 1024 > 0) {
    print $Channel $buffer;
  }
  close $Channel;
}

sub prep_fh {
  die "File name undefined" if !defined $File;
  die "File \"$File\" was not in lpsfile.map" if !exists $loadf{$File};

  my $msg = "";

  if (! $startpg and $File eq "lps_startpage.ps") {
    &log("Printing of start page was not requested, will be skipped");
    $msg = "ERROR=File not requested";
    print $Socket "\002101 $Seq %d %s", length $msg, $msg;
    return;
  }

 Alloc_Handle:
  for ($Handle=9; ; $Handle++) {
    if (!defined $fh[$Handle]) {
      &log("Created new handle \# $Handle") if $debug;
      $fh[$Handle] = return_fh() if !defined $fh[$Handle];
      last Alloc_Handle;
    }
    elsif (!defined fileno $fh[$Handle]) {
      &log("Re-using handle \# $Handle") if $debug;
      last Alloc_Handle;
    }
  }
  
  die "file handle \# $Handle already open" if defined fileno $fh[$Handle];
  my $binmode = "";
  $fn[$Handle] = $File;
  open $fh[$Handle], $loadf{$File} or 
    die "open file $File ($loadf{$File}): $!";
  if ($^O eq "MSWin32" and $Type eq "rb") {
    binmode $fh[$Handle];
    $binmode = " in binmode";
  }
  my $short_fn;
  ($short_fn = $loadf{$File}) =~ s{^$distro}{ROOT};
  &log("open $short_fn ($File)$binmode");

  $msg = sprintf "RETURN=$Handle";
  printf $Socket "\002101 $Seq %d %s", length $msg, $msg;
}

sub close_fh {
  &log("closing file handle") if $debug;
  close $fh[$Handle] or die "close $File (Handle: $Handle): $!";
  printf $Socket "\002101 $Seq 0 ";
}

sub log_msg {
  &log("logging msg") if $debug;
  print "$Log_txt\n";
}


sub acc_msg {
  open ACC, ">>$opts{a}";
  if ($? != 0){
    &log("Open Accounting $opts{a}: $!");
    return;
  }
  print ACC $Log_txt;
  close ACC;
}

sub error {
  &log("error ocurred: state: $State event: $Event");
}

sub nop {}

sub disconnect {
  &log("disconnecting from printer $lp");
  if (defined $Socket) {
    $s->remove($Socket);
    shutdown ($Socket, 2);# I have stopped using this socket
    close ($Socket);
    $Socket = undef;
  }
  $polled = 0;
}


sub exit_handler {
  &disconnect;
  die "\n";
}

sub return_fh {             # make anon filehandle
  local *FH;              # must be local, not my
  # now open it if you want to, then...
  return *FH;
}



sub get_event {
#  &log("get_event");
  my $pre_condition = shift;

  $Handle  = undef;
  $Length  = undef;
  $Type    = undef;
  $File    = undef;
  $Log_txt = undef;
  $Event   = undef;
  $Port    = undef;

  if ($pre_condition eq "lps_msg") {
    my ($resp, $first, $c, $stat);
    $stat = read $Socket, $first, 1;
    if (!defined $stat and $! == EAGAIN) {
      &log ("all requests from $lp handled") if $debug > 1;
      $Event = "";
      return "EAGAIN";
    }
    die "read: $! \n" if ! defined $stat;
    my $hex = ord $first;
    die "First char must be \\002, but was hex $hex\n" if $first ne "\002";
    
    my $header;
    my ($id, $len) = (undef, undef);
    
    ##
    ## Read header one char at a time as this is not a line oriented
    ## protocol. (read would wait for a \n)
    ##
  header_loop:
    while (read $Socket, $c, 1) {
      $header .= $c;
      
      if ($header =~ /^(\d+) (\d+) (\d+) $/) {
        ($id, $Seq, $len) = ($1, $2, $3);
        last header_loop ;
      }
    }
    &log("id: $id, seq: $Seq, len: $len") if $debug > 1;
    $stat = read $Socket, $resp, $len;
#    die "read: $!" if !defined $stat; 
    
    ## parse response and act accordingly
    if    ($id  == 42) {
      $Event = "snd_date";
    }
    elsif ($id  == 101 and 
           $resp =~ /SERVERID=(.*)\001PROTOCOL=(.*)\001KEEPALIVE=(\d+)/ ) {
      ## in this case, we received the printers revision, we 
      ## don't need to do anything.
      $Event = "ignore";
    }
    elsif ($id  == 0) {
      $Event = "ignore";
    }
    elsif ($id  == 51 and
           $resp =~ /^HANDLE=(\d+)\001COUNT=(\d+)\001LINE=(\d+)$/) {
      ## type r request. We need to send a text file.
      $Handle = $1;
      $Length = $2;
      $Type   = "r";
      $Event  = "data_req";
    }
    elsif ($id  == 51 and
           $resp =~ /^HANDLE=(\d+)\001COUNT=(\d+)$/) {
      ## type rb request. We need to send a binary file.
      $Handle = $1;
      $Length = $2;
      $Type   = "rb";
      $Event  = "data_req";
    }
    elsif ($id == 51 and 
           $resp =~ /^HANDLE=(\d+)\001COUNT=(\d+)\001OFFSET=(\d+)$/) {
      $Handle = $1;
      $Length = $2;
      $Type   = "rb";
      $Event  = "data_req";      
    }
    elsif ($id  == 53 and 
           $resp =~ /^HANDLE=(\d+)$/) {
      ## This is what the printer responds to the last message. (Makes us 
      ## close the file handle)
      $Handle = $1;
      $Event  = "fh_close";
    }
    elsif ($id  == 50 and
           $resp =~ /^PATH=(.*)\001TYPE=(.*)$/) {
      ## The printer requests a new file....
      $File   = $1;
      $Type   = $2;
      $Event  = "fh_req";
    }
    elsif ($id == 43) {
      ## Hmm, bit cheap, just treat accounting msgs as log messages
      $Log_txt= $resp;
      $Event  = "acc_msg";
    }
    elsif ($id == 44) {
      ## Hmm, bit cheap, just treat accounting msgs as log messages
      $Log_txt= $resp;
      $Event  = "log_msg";
    }
    elsif ($id  == 54 and 
           $resp =~ /^HANDLE=(\d+)\001PORT=(\d+)$/) {
      $Handle = $1;
      $Port   = $2;
      $Event  = "snd_high";
    }
    else {
      &log("unsolicited input received: id: $id, seq: $Seq, data: $resp");
      $Event  = "erred";
    }
  }
  elsif ($pre_condition eq "timeout") {
    $Event = &poll_lp; #poll_lp returns lp_up or lp_down
  }
  else {
    die "invalid pre_condition: \"$pre_condition\"\n";
  }
#  &log("got event \"$Event\"") if $debug;
  
}

sub long_to_short {
  my ($h_ref) = @_;
  my $map     = $distro . "/servers/lpsfile.map";
  my $missing = 0;
  open MAP, $map or die "open file $map: $!";
 map_loop:
  while (<MAP>) {
    s/\s*#.*//g;
    next map_loop if /^$/;
    my ($key, $val) = split (/\s+/);
    if (-f "$distro/servers/$val") {
      $$h_ref{$key}="$distro/servers/$val";
    }
    elsif (-f "$distro/servers/sys/$val") {
      $$h_ref{$key}="$distro/servers/sys/$val";
    }
    else {
      &log("unable to locate file \"$key\"") if $debug;
      $missing += 1;
    }
  }
  close MAP;
  &log("There are $missing files not in lpsfile.map") if $missing > 0;

  ## On top of all these files, there are some more requests...

  my %spec = ('$CONFIG'      => "/servers/$lp.cfg",
              '$DEFAULTS'    => "/servers/GENERIC.dft",
              '$SYSPARAMS'   => "/servers/GENERIC.spf",
              '$PCLDEFAULTS' => "/servers/GENERIC.pdf",
              '$RESOURCES'   => "/servers/GENERIC.res",
              'ansipro.ps'   => "/servers/ansipro.ps", 
              'ps_supp.ps'   => "/servers/ps_supp.ps", 
              '$SETUP'       => "/servers/GENERIC.sup");

  foreach my $k (keys %spec) {
    if (-f $distro . $spec{$k}) {
      $$h_ref{$k} = $distro . $spec{$k};
    }
    else {
      die "Could not find $k file at $distro . $spec{$k}";
    }
  }
}


sub mk_printercfg {
  my $time = localtime;
  my $actg = exists $opts{a} ? "on"     : "off";
  my $nmask= exists $opts{n} ? $opts{n} : "255.255.255.128";

  die "$distro is missing files\n" if (!-d "$distro/servers");

  open  CFG, ">$distro/servers/$lp.cfg" or die "mk $lp.cfg: $!";
  print CFG<<EO_CONFIG;
# Configuration file for PrintServer printer "$lp"
# Created by $0 on $time
# WARNING: Do NOT modify without first consulting the LPS documentation!
# Note:  All configuration keywords must be in lowercase for proper
#        parsing at a later time by the lpssetup script!
#####

printername $lp
password (LPS)
language eng
accounting $actg
logging on
subnetmask $nmask
option snmp
snmpenableauthentraps off
community public 0.0.0.0 read
EO_CONFIG
  close CFG;
}

sub log {
  my ($txt) = @_;
  my $date  = localtime;
  ($date)   = ($date =~ /^... (...\s+\d{1,2} ..:..:..)/);
  printf "$date: %s\n", $txt;
}

sub usage {
  my $this_file = basename ($0);

  print <<EO_USAGE;

  $this_file [-hvsb] [-a /path/to/acc-file] [-n netmask] -d /path 
  [-c /path/to/bootptab] printer_name

  use this program to download boot files to a DEC Printserver 17 (aka lps17)

  options:
  -h print this info
  -H print more info
  -v print debug information
  -t print trace messages. This is gonna flood you with messages.
  -d directory containing DEC (Compaq) printer SW, see note 1)
  -a turns on accounting. 
  -n net mask. This will be passed to the printer
  -b Do not use built in bootp daemon.
  -c bootptab style config file.
  -s print startpage
  -r run as detached daemon. Messages will be written to $this_file.log.
     This is not supported on Windows.
  -l use given file as log file

  example: 
  $this_file -v -d F:/Downloads/lps17/LPS_SOL_BIN_V052 \\
    -c F:/Downloads/lps17/bootptab.txt 160.59.218.20

 Note:
 1) You may download this at: www.digital.com/printers

EO_USAGE
  die "\n";

}

sub podthis {
  $^W = 0;
  pod2text $0;
  die "\n";

}
